/** @name apprsfx.cpp Extracts The Apprentice sound effects from level data.
 *
 * @author CD-i Fan <cdifan@gmail.com>
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

//* Data sector size.
#define SECTOR_SIZE 2048

//* Rounds to integral sector size.
#define round_sector(n) \
	(((n) + SECTOR_SIZE - 1) & ~(SECTOR_SIZE - 1))

//* Gets byte (8-bit) value from file.
#define getbyte(f) \
	((unsigned char)getc(f))

//* Gets short (16-bit) value from file.
short getshort(FILE *f)
{
	short n = (short)getbyte(f) << 8;
	n |= (short)getbyte(f);

	return n;
}

//* Gets long (32-bit) value from file.
long getlong(FILE *f)
{
	long n = (long)getbyte(f) << 24;
	n |= (long)getbyte(f) << 16;
	n |= (long)getbyte(f) << 8;
	n |= (long)getbyte(f);

	return n;
}

//* Puts long (32-bit) value to file.
void putlong(long n, FILE *f)
{
	putc((n >> 24) & 0xFF, f);
	putc((n >> 16) & 0xFF, f);
	putc((n >> 8) & 0xFF, f);
	putc(n & 0xFF, f);
}

//* Gets block position.
long blockpos(long *blocksize, int iblock)
{
	long pos = SECTOR_SIZE;
	while (iblock-- > 0)
		pos += round_sector(*blocksize++);

	return pos;
}

//* Extract sound effects from level data file.
int extractsfx(const char *level)
{
	// Open level data file.
	FILE *flevel = fopen(level, "rb");
	if (flevel == NULL)
	{
		fprintf(stderr, "%s: Cannot open\n", level);
		return 1;
	}

	const char *levelnump = level;
	while (*levelnump != '\0' && !isdigit(*levelnump))
		levelnump++;

	int levelnum = atoi(levelnump);

	printf("Extracting The Apprentice sound effects from level %d data:\n", levelnum);

	printf("Reading block header from %s @$0...\n", level);

	// Read block count and list of block lengths.
	short nblocks = getshort(flevel);
	long *blocksize = new long[nblocks];
	for (short iblock = 0; iblock < nblocks; iblock++)
		blocksize[iblock] = getlong(flevel);

	// Sound data is in block 4.
	long soundpos = blockpos(blocksize, 4);
	fseek(flevel, soundpos, SEEK_SET);
	
	printf("Reading sound header from %s @$%lx...\n", level, soundpos);

	// Read effect count and effect position, groups and channel.
	long neffects = getlong(flevel);
	long audiotype = getlong(flevel);
	long *effectpos = new long[neffects];
	short *effectgrps = new short[neffects];
	unsigned char *effectchan = new unsigned char[neffects];
	for (short ieffect = 1; ieffect < neffects; ++ieffect)
	{
		long pos = ftell(flevel);
		effectpos[ieffect] = pos + getlong(flevel);
		effectgrps[ieffect] = getshort(flevel);
		effectchan[ieffect] = getbyte(flevel);
		getbyte(flevel);
	}

	// Extract each effect in turn.
	for (short ieffect = 1; ieffect < neffects; ++ieffect)
	{
		// Skip effects with length zero.
		if (effectgrps[ieffect] == 0)
			continue;

		// Construct effect file name.
		char effect[16];
		sprintf(effect, "effect%d_%d.raw", levelnum, ieffect);

		// Create effect file.
		FILE *feffect = fopen(effect, "wb");
		if (feffect == NULL)
		{
			fprintf(stderr, "%s: Cannot create\n", effect);
			continue;
		}

		printf("Extracting %s (%d groups) from %s @$%lx...\n", effect, effectgrps[ieffect], level, effectpos[ieffect]);

		// Extract effect data.
		fseek(flevel, effectpos[ieffect], SEEK_SET);
		for (int igrp = 0; igrp < effectgrps[ieffect]; ++igrp)
		{
			// Expand 8 sound parameter bytes to 16, masking and shifting as required.
			for (int iparam = 0; iparam < 2; ++iparam)
			{
				long param = getlong(flevel);
				if (effectchan[ieffect] == 0)
					param >>= 8;
				param &= 0x00FF00FF;
				putlong(param, feffect);
				putlong(param, feffect);
			}

			// Mask and shift 112 sound data bytes as required.
			for (int idata = 0; idata < 28; ++idata)
			{
				long data = getlong(flevel);
				if (effectchan[ieffect] == 0)
					data <<= 4;
				data &= 0xF0F0F0F0;
				putlong(data, feffect);
			}
		}

		fclose(feffect);
	}

	fclose(flevel);

	return 0;
}

int main(int argc, const char **argv)
{
	if (argc != 2)
	{
		fprintf(stderr, "Usage: apprsfx levelN.dat\n");
		fprintf(stderr, "Function: Extracts The Apprentice sound effects from level data\n");
		fprintf(stderr, "Raw sound effect data is written to effectN_M.raw\n");
		fprintf(stderr, "Written by CD-i Fan <cdifan@gmail.com>\n");
		exit(2);
	}

	return extractsfx(argv[1]);
}

